@@ -35,7 +35,7 @@
 #include <ctype.h>          /* isalpha */
 
 /* misc win32 helpers */
-#ifdef WIN32
+#ifdef _WIN32
 # ifndef HAVE_GETTIMEOFDAY
    /* replacement gettimeofday implementation */
 #  include <sys/timeb.h>
@@ -79,7 +79,7 @@ static inline int _private_gettimeofday( struct timeval *tv, void *tz )
 
 #define DEFAULT_UDF_CACHE_LEVEL 1
 
-struct dvd_reader_s {
+struct dvd_reader_device_s {
   /* Basic information. */
   int isImageFile;
 
@@ -103,7 +103,7 @@ struct dvd_reader_s {
 
 struct dvd_file_s {
   /* Basic information. */
-  dvd_reader_t *dvd;
+  dvd_reader_t *ctx;
 
   /* Hack for selecting the right css title. */
   int css_title;
@@ -129,9 +129,9 @@ struct dvd_file_s {
  * level = 0 (no caching)
  * level = 1 (caching filesystem info)
  */
-int DVDUDFCacheLevel(dvd_reader_t *device, int level)
+int DVDUDFCacheLevel(dvd_reader_t *reader, int level)
 {
-  struct dvd_reader_s *dev = (struct dvd_reader_s *)device;
+  dvd_reader_device_t *dev = reader->rd;
 
   if(level > 0) {
     level = 1;
@@ -144,16 +144,16 @@ int DVDUDFCacheLevel(dvd_reader_t *device, int level)
   return level;
 }
 
-void *GetUDFCacheHandle(dvd_reader_t *device)
+void *GetUDFCacheHandle(dvd_reader_t *reader)
 {
-  struct dvd_reader_s *dev = (struct dvd_reader_s *)device;
+  dvd_reader_device_t *dev = reader->rd;
 
   return dev->udfcache;
 }
 
-void SetUDFCacheHandle(dvd_reader_t *device, void *cache)
+void SetUDFCacheHandle(dvd_reader_t *reader, void *cache)
 {
-  struct dvd_reader_s *dev = (struct dvd_reader_s *)device;
+  dvd_reader_device_t *dev = reader->rd;
 
   dev->udfcache = cache;
 }
@@ -161,8 +161,9 @@ void SetUDFCacheHandle(dvd_reader_t *device, void *cache)
 
 
 /* Loop over all titles and call dvdcss_title to crack the keys. */
-static int initAllCSSKeys( dvd_reader_t *dvd )
+static int initAllCSSKeys( dvd_reader_t *ctx )
 {
+  dvd_reader_device_t *dvd = ctx->rd;
   struct timeval all_s, all_e;
   struct timeval t_s, t_e;
   char filename[ MAX_UDF_FILE_NAME_LEN ];
@@ -173,10 +174,8 @@ static int initAllCSSKeys( dvd_reader_t *dvd )
   if(nokeys_str != NULL)
     return 0;
 
-  fprintf( stderr, "\n" );
-  fprintf( stderr, "libdvdread: Attempting to retrieve all CSS keys\n" );
-  fprintf( stderr, "libdvdread: This can take a _long_ time, "
-           "please be patient\n\n" );
+  Log2(ctx,"Attempting to retrieve all CSS keys" );
+  Log2(ctx,"This can take a _long_ time, please be patient" );
   gettimeofday(&all_s, NULL);
 
   for( title = 0; title < 100; title++ ) {
@@ -186,42 +185,37 @@ static int initAllCSSKeys( dvd_reader_t *dvd )
     } else {
       sprintf( filename, "/VIDEO_TS/VTS_%02d_%d.VOB", title, 0 );
     }
-    start = UDFFindFile( dvd, filename, &len );
+    start = UDFFindFile( ctx, filename, &len );
     if( start != 0 && len != 0 ) {
       /* Perform CSS key cracking for this title. */
-      fprintf( stderr, "libdvdread: Get key for %s at 0x%08x\n",
-               filename, start );
+      Log3(ctx,"Get key for %s at 0x%08x",filename, start );
       if( dvdinput_title( dvd->dev, (int)start ) < 0 ) {
-        fprintf( stderr, "libdvdread: Error cracking CSS key for %s (0x%08x)\n", filename, start);
+        Log1(ctx,"Error cracking CSS key for %s (0x%08x)", filename, start);
       }
       gettimeofday( &t_e, NULL );
-      fprintf( stderr, "libdvdread: Elapsed time %ld\n",
-               (long int) t_e.tv_sec - t_s.tv_sec );
+      Log3(ctx,"Elapsed time %ld", (long int) t_e.tv_sec - t_s.tv_sec );
     }
 
     if( title == 0 ) continue;
 
     gettimeofday( &t_s, NULL );
     sprintf( filename, "/VIDEO_TS/VTS_%02d_%d.VOB", title, 1 );
-    start = UDFFindFile( dvd, filename, &len );
+    start = UDFFindFile( ctx, filename, &len );
     if( start == 0 || len == 0 ) break;
 
     /* Perform CSS key cracking for this title. */
-    fprintf( stderr, "libdvdread: Get key for %s at 0x%08x\n",
-             filename, start );
+    Log3(ctx,"Get key for %s at 0x%08x",filename, start );
     if( dvdinput_title( dvd->dev, (int)start ) < 0 ) {
-      fprintf( stderr, "libdvdread: Error cracking CSS key for %s (0x%08x)!!\n", filename, start);
+      Log1(ctx,"Error cracking CSS key for %s (0x%08x)", filename, start);
     }
     gettimeofday( &t_e, NULL );
-    fprintf( stderr, "libdvdread: Elapsed time %ld\n",
-             (long int) t_e.tv_sec - t_s.tv_sec );
+    Log3(ctx,"Elapsed time %ld", (long int) t_e.tv_sec - t_s.tv_sec );
   }
   title--;
 
-  fprintf( stderr, "libdvdread: Found %d VTS's\n", title );
+  Log3(ctx,"Found %d VTS's", title );
   gettimeofday(&all_e, NULL);
-  fprintf( stderr, "libdvdread: Elapsed time %ld\n",
-           (long int) all_e.tv_sec - all_s.tv_sec );
+  Log3(ctx,"Elapsed time %ld", (long int) all_e.tv_sec - all_s.tv_sec );
 
   return 0;
 }
@@ -231,21 +225,21 @@ static int initAllCSSKeys( dvd_reader_t *dvd )
 /**
  * Open a DVD image or block device file or use stream_cb functions.
  */
-static dvd_reader_t *DVDOpenImageFile( const char *location,
-                                       void *stream,
-                                       dvd_reader_stream_cb *stream_cb,
+static dvd_reader_device_t *DVDOpenImageFile( dvd_reader_t *ctx,
+                                              const char *location,
+                                        dvd_reader_stream_cb *stream_cb,
                                        int have_css )
 {
-  dvd_reader_t *dvd;
+  dvd_reader_device_t *dvd;
   dvd_input_t dev;
 
-  dev = dvdinput_open( location, stream, stream_cb );
+  dev = dvdinput_open( ctx->priv, &ctx->logcb, location, stream_cb );
   if( !dev ) {
-    fprintf( stderr, "libdvdread: Can't open %s for reading\n", location );
+    Log0(ctx,"Can't open %s for reading", location );
     return NULL;
   }
 
-  dvd = calloc( 1, sizeof( dvd_reader_t ) );
+  dvd = calloc( 1, sizeof( dvd_reader_device_t ) );
   if( !dvd ) {
     dvdinput_close(dev);
     return NULL;
@@ -266,11 +260,11 @@ static dvd_reader_t *DVDOpenImageFile( const char *location,
   return dvd;
 }
 
-static dvd_reader_t *DVDOpenPath( const char *path_root )
+static dvd_reader_device_t *DVDOpenPath( const char *path_root )
 {
-  dvd_reader_t *dvd;
+  dvd_reader_device_t *dvd;
 
-  dvd = calloc( 1, sizeof( dvd_reader_t ) );
+  dvd = calloc( 1, sizeof( dvd_reader_device_t ) );
   if( !dvd ) return NULL;
   dvd->path_root = strdup( path_root );
   if(!dvd->path_root) {
@@ -336,25 +330,38 @@ static char *bsd_block2char( const char *path )
 }
 #endif
 
-static dvd_reader_t *DVDOpenCommon( const char *ppath,
-                                    void *stream,
+static dvd_reader_t *DVDOpenCommon( void *priv,
+                                    const dvd_logger_cb *logcb,
+                                    const char *ppath,
                                     dvd_reader_stream_cb *stream_cb )
 {
   struct stat fileinfo;
   int ret, have_css, retval, cdir = -1;
-  dvd_reader_t *ret_val = NULL;
   char *dev_name = NULL;
   char *path = NULL, *new_path = NULL, *path_copy = NULL;
+  dvd_reader_t *ctx = calloc(1, sizeof(*ctx));
+  if(!ctx)
+      return NULL;
+
+  ctx->priv = priv;
+  if(logcb)
+    ctx->logcb = *logcb;
 
 #if defined(_WIN32) || defined(__OS2__)
       int len;
 #endif
 
   /* Try to open DVD using stream_cb functions */
-  if( stream != NULL && stream_cb != NULL )
+  if( priv != NULL && stream_cb != NULL )
   {
-    have_css = dvdinput_setup();
-    return DVDOpenImageFile( NULL, stream, stream_cb, have_css );
+    have_css = dvdinput_setup( ctx->priv, &ctx->logcb );
+    ctx->rd = DVDOpenImageFile( ctx, NULL, stream_cb, have_css );
+    if(!ctx->rd)
+    {
+        free(ctx);
+        return NULL;
+    }
+    return ctx;
   }
 
   if( ppath == NULL )
@@ -365,7 +372,7 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
     goto DVDOpen_error;
 
   /* Try to open libdvdcss or fall back to standard functions */
-  have_css = dvdinput_setup();
+  have_css = dvdinput_setup( ctx->priv, &ctx->logcb );
 
 #if defined(_WIN32) || defined(__OS2__)
   /* Strip off the trailing \ if it is not a drive */
@@ -384,13 +391,18 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
 
     /* maybe "host:port" url? try opening it with acCeSS library */
     if( strchr(path,':') ) {
-      ret_val = DVDOpenImageFile( path, NULL, NULL, have_css );
+      ctx->rd = DVDOpenImageFile( ctx, path, NULL, have_css );
       free(path);
-      return ret_val;
+      if(!ctx->rd)
+      {
+          free(ctx);
+          return NULL;
+      }
+      return ctx;
     }
 
     /* If we can't stat the file, give up */
-    fprintf( stderr, "libdvdread: Can't stat %s\n", path );
+    Log0(ctx, "Can't stat %s", path );
     perror("");
     goto DVDOpen_error;
   }
@@ -403,7 +415,6 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
     /**
      * Block devices and regular files are assumed to be DVD-Video images.
      */
-    dvd_reader_t *dvd = NULL;
 #if defined(__sun)
     dev_name = sun_block2char( path );
 #elif defined(SYS_BSD)
@@ -413,12 +424,16 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
 #endif
     if(!dev_name)
         goto DVDOpen_error;
-    dvd = DVDOpenImageFile( dev_name, NULL, NULL, have_css );
+    ctx->rd = DVDOpenImageFile( ctx, dev_name, NULL, have_css );
     free( dev_name );
     free(path);
-    return dvd;
+    if(!ctx->rd)
+    {
+        free(ctx);
+        return NULL;
+    }
+    return ctx;
   } else if( S_ISDIR( fileinfo.st_mode ) ) {
-    dvd_reader_t *auth_drive = 0;
 #if defined(SYS_BSD)
     struct fstab* fe;
 #elif defined(__sun) || defined(__linux__)
@@ -429,7 +444,7 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
     if( !(path_copy = strdup( path ) ) )
       goto DVDOpen_error;
 
-#ifndef WIN32 /* don't have fchdir, and getcwd( NULL, ... ) is strange */
+#ifndef _WIN32 /* don't have fchdir, and getcwd( NULL, ... ) is strange */
               /* Also WIN32 does not have symlinks, so we don't need this bit of code. */
 
     /* Resolve any symlinks and get the absolute dir name. */
@@ -500,12 +515,11 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
         for (i=0; i<r; i++) {
             if (!strcmp(path_copy, s[i].f_mntonname)) {
                 dev_name = bsd_block2char(s[i].f_mntfromname);
-                fprintf( stderr,
-                        "libdvdread: Attempting to use device %s"
-                        " mounted on %s for CSS authentication\n",
+                Log3(ctx, "Attempting to use device %s"
+                          " mounted on %s for CSS authentication",
                         dev_name,
                         s[i].f_mntonname);
-                auth_drive = DVDOpenImageFile( dev_name, NULL, NULL, have_css );
+                ctx->rd = DVDOpenImageFile( ctx, dev_name, NULL, have_css );
                 break;
             }
         }
@@ -513,12 +527,11 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
 #elif defined(SYS_BSD)
     if( ( fe = getfsfile( path_copy ) ) ) {
       dev_name = bsd_block2char( fe->fs_spec );
-      fprintf( stderr,
-               "libdvdread: Attempting to use device %s"
-               " mounted on %s for CSS authentication\n",
+      Log3(ctx, "Attempting to use device %s"
+               " mounted on %s for CSS authentication",
                dev_name,
                fe->fs_file );
-      auth_drive = DVDOpenImageFile( dev_name, NULL, NULL, have_css );
+      ctx->rd = DVDOpenImageFile( ctx, dev_name, NULL, have_css );
     }
 #elif defined(__sun)
     mntfile = fopen( MNTTAB, "r" );
@@ -529,12 +542,11 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
       while( ( res = getmntent( mntfile, &mp ) ) != -1 ) {
         if( res == 0 && !strcmp( mp.mnt_mountp, path_copy ) ) {
           dev_name = sun_block2char( mp.mnt_special );
-          fprintf( stderr,
-                   "libdvdread: Attempting to use device %s"
-                   " mounted on %s for CSS authentication\n",
+          Log3(ctx, "Attempting to use device %s"
+                   " mounted on %s for CSS authentication",
                    dev_name,
                    mp.mnt_mountp );
-          auth_drive = DVDOpenImageFile( dev_name, NULL, NULL, have_css );
+          ctx->rd = DVDOpenImageFile( ctx, dev_name, NULL, have_css );
           break;
         }
       }
@@ -543,16 +555,21 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
 #elif defined(__linux__)
     mntfile = fopen( _PATH_MOUNTED, "r" );
     if( mntfile ) {
-      struct mntent *me;
 
+#ifdef HAVE_GETMNTENT_R
+      struct mntent *me, mbuf;
+      char buf [8192];
+      while( ( me = getmntent_r( mntfile, &mbuf, buf, sizeof(buf) ) ) ) {
+#else
+      struct mntent *me;
       while( ( me = getmntent( mntfile ) ) ) {
+#endif
         if( !strcmp( me->mnt_dir, path_copy ) ) {
-          fprintf( stderr,
-                   "libdvdread: Attempting to use device %s"
-                   " mounted on %s for CSS authentication\n",
+          Log3(ctx, "Attempting to use device %s"
+                   " mounted on %s for CSS authentication",
                    me->mnt_fsname,
                    me->mnt_dir );
-          auth_drive = DVDOpenImageFile( me->mnt_fsname, NULL, NULL, have_css );
+          ctx->rd = DVDOpenImageFile( ctx, me->mnt_fsname, NULL, have_css );
           dev_name = strdup(me->mnt_fsname);
           break;
         }
@@ -566,20 +583,20 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
         ( !path[2] ||
           ((path[2] == '\\' || path[2] == '/') && !path[3])))
 #endif
-    auth_drive = DVDOpenImageFile( path, NULL, NULL, have_css );
+    ctx->rd = DVDOpenImageFile( ctx, path, NULL, have_css );
 #endif
 
 #if !defined(_WIN32) && !defined(__OS2__)
     if( !dev_name ) {
-      fprintf( stderr, "libdvdread: Couldn't find device name.\n" );
-    } else if( !auth_drive ) {
-      fprintf( stderr, "libdvdread: Device %s inaccessible, "
-               "CSS authentication not available.\n", dev_name );
+      Log0(ctx, "Couldn't find device name." );
+    } else if( !ctx->rd ) {
+      Log0(ctx, "Device %s inaccessible, "
+                "CSS authentication not available.", dev_name );
     }
 #else
-    if( !auth_drive ) {
-        fprintf( stderr, "libdvdread: Device %s inaccessible, "
-                 "CSS authentication not available.\n", path );
+    if( !ctx->rd ) {
+        Log0(ctx, "Device %s inaccessible, "
+                 "CSS authentication not available.", path );
     }
 #endif
 
@@ -591,21 +608,27 @@ static dvd_reader_t *DVDOpenCommon( const char *ppath,
     /**
      * If we've opened a drive, just use that.
      */
-    if( auth_drive ) {
-      free(path);
-      return auth_drive;
+    if(ctx->rd)
+    {
+        free(path);
+        return ctx;
     }
     /**
      * Otherwise, we now try to open the directory tree instead.
      */
-    ret_val = DVDOpenPath( path );
-      free( path );
-      return ret_val;
+    ctx->rd = DVDOpenPath( path );
+    free( path );
+    if(!ctx->rd)
+    {
+        free(ctx);
+        return NULL;
+    }
+    return ctx;
   }
 
 DVDOpen_error:
   /* If it's none of the above, screw it. */
-  fprintf( stderr, "libdvdread: Could not open %s\n", path );
+  Log0(ctx, "Could not open %s", path );
   free( path );
   free( path_copy );
   if ( cdir >= 0 )
@@ -616,21 +639,34 @@ DVDOpen_error:
 
 dvd_reader_t *DVDOpen( const char *ppath )
 {
-    return DVDOpenCommon( ppath, NULL, NULL );
+    return DVDOpenCommon( NULL, NULL, ppath, NULL );
 }
 
 dvd_reader_t *DVDOpenStream( void *stream,
                              dvd_reader_stream_cb *stream_cb )
 {
-    return DVDOpenCommon( NULL, stream, stream_cb );
+    return DVDOpenCommon( stream, NULL, NULL, stream_cb );
+}
+
+dvd_reader_t *DVDOpen2( void *priv, const dvd_logger_cb *logcb,
+                        const char *ppath )
+{
+    return DVDOpenCommon( priv, logcb, ppath, NULL );
+}
+
+dvd_reader_t *DVDOpenStream2( void *priv, const dvd_logger_cb *logcb,
+                              dvd_reader_stream_cb *stream_cb )
+{
+    return DVDOpenCommon( priv, logcb, NULL, stream_cb );
 }
 
 void DVDClose( dvd_reader_t *dvd )
 {
   if( dvd ) {
-    if( dvd->dev ) dvdinput_close( dvd->dev );
-    if( dvd->path_root ) free( dvd->path_root );
-    if( dvd->udfcache ) FreeUDFCache( dvd->udfcache );
+    if( dvd->rd->dev ) dvdinput_close( dvd->rd->dev );
+    if( dvd->rd->path_root ) free( dvd->rd->path_root );
+    if( dvd->rd->udfcache ) FreeUDFCache( dvd->rd->udfcache );
+    free( dvd->rd );
     free( dvd );
   }
 }
@@ -638,24 +674,24 @@ void DVDClose( dvd_reader_t *dvd )
 /**
  * Open an unencrypted file on a DVD image file.
  */
-static dvd_file_t *DVDOpenFileUDF( dvd_reader_t *dvd, const char *filename,
+static dvd_file_t *DVDOpenFileUDF( dvd_reader_t *ctx, const char *filename,
                                    int do_cache )
 {
   uint32_t start, len;
   dvd_file_t *dvd_file;
 
-  start = UDFFindFile( dvd, filename, &len );
+  start = UDFFindFile( ctx, filename, &len );
   if( !start ) {
-    fprintf( stderr, "libdvdread:DVDOpenFileUDF:UDFFindFile %s failed\n", filename );
+    Log0(ctx, "DVDOpenFileUDF:UDFFindFile %s failed", filename );
     return NULL;
   }
 
   dvd_file = calloc( 1, sizeof( dvd_file_t ) );
   if( !dvd_file ) {
-    fprintf( stderr, "libdvdread:DVDOpenFileUDF:malloc failed\n" );
+    Log0(ctx, "DVDOpenFileUDF:malloc failed" );
     return NULL;
   }
-  dvd_file->dvd = dvd;
+  dvd_file->ctx = ctx;
   dvd_file->lb_start = start;
   dvd_file->filesize = len / DVD_VIDEO_LB_LEN;
 
@@ -668,7 +704,7 @@ static dvd_file_t *DVDOpenFileUDF( dvd_reader_t *dvd, const char *filename,
     if( !dvd_file->cache )
         return dvd_file;
 
-    ret = InternalUDFReadBlocksRaw( dvd, dvd_file->lb_start,
+    ret = InternalUDFReadBlocksRaw( ctx, dvd_file->lb_start,
                                     dvd_file->filesize, dvd_file->cache,
                                     DVDINPUT_NOFLAGS );
     if( ret != dvd_file->filesize ) {
@@ -719,16 +755,16 @@ static int findDVDFile( dvd_reader_t *dvd, const char *file, char *filename )
     nodirfile = file;
   }
 
-  ret = findDirFile( dvd->path_root, nodirfile, filename );
+  ret = findDirFile( dvd->rd->path_root, nodirfile, filename );
   if( ret < 0 ) {
     char video_path[ PATH_MAX + 1 ];
 
     /* Try also with adding the path, just in case. */
-    sprintf( video_path, "%s/VIDEO_TS/", dvd->path_root );
+    sprintf( video_path, "%s/VIDEO_TS/", dvd->rd->path_root );
     ret = findDirFile( video_path, nodirfile, filename );
     if( ret < 0 ) {
       /* Try with the path, but in lower case. */
-      sprintf( video_path, "%s/video_ts/", dvd->path_root );
+      sprintf( video_path, "%s/video_ts/", dvd->rd->path_root );
       ret = findDirFile( video_path, nodirfile, filename );
       if( ret < 0 ) {
         return 0;
@@ -742,7 +778,7 @@ static int findDVDFile( dvd_reader_t *dvd, const char *file, char *filename )
 /**
  * Open an unencrypted file from a DVD directory tree.
  */
-static dvd_file_t *DVDOpenFilePath( dvd_reader_t *dvd, const char *filename )
+static dvd_file_t *DVDOpenFilePath( dvd_reader_t *ctx, const char *filename )
 {
   char full_path[ PATH_MAX + 1 ];
   dvd_file_t *dvd_file;
@@ -750,27 +786,27 @@ static dvd_file_t *DVDOpenFilePath( dvd_reader_t *dvd, const char *filename )
   dvd_input_t dev;
 
   /* Get the full path of the file. */
-  if( !findDVDFile( dvd, filename, full_path ) ) {
-    fprintf( stderr, "libdvdread:DVDOpenFilePath:findDVDFile %s failed\n", filename );
+  if( !findDVDFile( ctx, filename, full_path ) ) {
+    Log0(ctx, "DVDOpenFilePath:findDVDFile %s failed", filename );
     return NULL;
   }
 
-  dev = dvdinput_open( full_path, NULL, NULL );
+  dev = dvdinput_open( ctx->priv, &ctx->logcb, full_path, NULL );
   if( !dev ) {
-    fprintf( stderr, "libdvdread:DVDOpenFilePath:dvdinput_open %s failed\n", full_path );
+    Log0(ctx, "DVDOpenFilePath:dvdinput_open %s failed", full_path );
     return NULL;
   }
 
   dvd_file = calloc( 1, sizeof( dvd_file_t ) );
   if( !dvd_file ) {
-    fprintf( stderr, "libdvdread:DVDOpenFilePath:dvd_file malloc failed\n" );
+    Log0(ctx, "DVDOpenFilePath:dvd_file malloc failed" );
     dvdinput_close(dev);
     return NULL;
   }
-  dvd_file->dvd = dvd;
+  dvd_file->ctx = ctx;
 
   if( stat( full_path, &fileinfo ) < 0 ) {
-    fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
+    Log0(ctx, "Can't stat() %s.", filename );
     free( dvd_file );
     dvdinput_close( dev );
     return NULL;
@@ -782,7 +818,7 @@ static dvd_file_t *DVDOpenFilePath( dvd_reader_t *dvd, const char *filename )
   return dvd_file;
 }
 
-static dvd_file_t *DVDOpenVOBUDF( dvd_reader_t *dvd, int title, int menu )
+static dvd_file_t *DVDOpenVOBUDF( dvd_reader_t *ctx, int title, int menu )
 {
   char filename[ MAX_UDF_FILE_NAME_LEN ];
   uint32_t start, len;
@@ -793,12 +829,12 @@ static dvd_file_t *DVDOpenVOBUDF( dvd_reader_t *dvd, int title, int menu )
   } else {
     sprintf( filename, "/VIDEO_TS/VTS_%02d_%d.VOB", title, menu ? 0 : 1 );
   }
-  start = UDFFindFile( dvd, filename, &len );
+  start = UDFFindFile( ctx, filename, &len );
   if( start == 0 ) return NULL;
 
   dvd_file = calloc( 1, sizeof( dvd_file_t ) );
   if( !dvd_file ) return NULL;
-  dvd_file->dvd = dvd;
+  dvd_file->ctx = ctx;
   /*Hack*/ dvd_file->css_title = title << 1 | menu;
   dvd_file->lb_start = start;
   dvd_file->filesize = len / DVD_VIDEO_LB_LEN;
@@ -809,26 +845,25 @@ static dvd_file_t *DVDOpenVOBUDF( dvd_reader_t *dvd, int title, int menu )
 
     for( cur = 2; cur < 10; cur++ ) {
       sprintf( filename, "/VIDEO_TS/VTS_%02d_%d.VOB", title, cur );
-      if( !UDFFindFile( dvd, filename, &len ) ) break;
+      if( !UDFFindFile( ctx, filename, &len ) ) break;
       dvd_file->filesize += len / DVD_VIDEO_LB_LEN;
     }
   }
 
-  if( dvd->css_state == 1 /* Need key init */ ) {
-    initAllCSSKeys( dvd );
-    dvd->css_state = 2;
+  if( ctx->rd->css_state == 1 /* Need key init */ ) {
+    initAllCSSKeys( ctx );
+    ctx->rd->css_state = 2;
   }
   /*
   if( dvdinput_title( dvd_file->dvd->dev, (int)start ) < 0 ) {
-      fprintf( stderr, "libdvdread: Error cracking CSS key for %s\n",
-               filename );
+      Log0(ctx, "Error cracking CSS key for %s", filename );
   }
   */
 
   return dvd_file;
 }
 
-static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *dvd, int title, int menu )
+static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *ctx, int title, int menu )
 {
   char filename[ MAX_UDF_FILE_NAME_LEN ];
   char full_path[ PATH_MAX + 1 ];
@@ -837,7 +872,7 @@ static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *dvd, int title, int menu )
 
   dvd_file = calloc( 1, sizeof( dvd_file_t ) );
   if( !dvd_file ) return NULL;
-  dvd_file->dvd = dvd;
+  dvd_file->ctx = ctx;
   /*Hack*/ dvd_file->css_title = title << 1 | menu;
 
   if( menu ) {
@@ -848,19 +883,19 @@ static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *dvd, int title, int menu )
     } else {
       sprintf( filename, "VTS_%02i_0.VOB", title );
     }
-    if( !findDVDFile( dvd, filename, full_path ) ) {
+    if( !findDVDFile( ctx, filename, full_path ) ) {
       free( dvd_file );
       return NULL;
     }
 
-    dev = dvdinput_open( full_path, NULL, NULL );
+    dev = dvdinput_open( ctx->priv, &ctx->logcb, full_path, NULL );
     if( dev == NULL ) {
       free( dvd_file );
       return NULL;
     }
 
     if( stat( full_path, &fileinfo ) < 0 ) {
-      fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
+      Log0(ctx, "Can't stat() %s.", filename );
       dvdinput_close(dev);
       free( dvd_file );
       return NULL;
@@ -876,17 +911,17 @@ static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *dvd, int title, int menu )
     for( i = 0; i < TITLES_MAX; ++i ) {
 
       sprintf( filename, "VTS_%02i_%i.VOB", title, i + 1 );
-      if( !findDVDFile( dvd, filename, full_path ) ) {
+      if( !findDVDFile( ctx, filename, full_path ) ) {
         break;
       }
 
       if( stat( full_path, &fileinfo ) < 0 ) {
-        fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
+        Log0(ctx, "Can't stat() %s.", filename );
         break;
       }
 
       dvd_file->title_sizes[ i ] = fileinfo.st_size / DVD_VIDEO_LB_LEN;
-      dvd_file->title_devs[ i ] = dvdinput_open( full_path, NULL, NULL );
+      dvd_file->title_devs[ i ] = dvdinput_open( ctx->priv, &ctx->logcb, full_path, NULL );
       dvdinput_title( dvd_file->title_devs[ i ], 0 );
       dvd_file->filesize += dvd_file->title_sizes[ i ];
     }
@@ -899,9 +934,10 @@ static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *dvd, int title, int menu )
   return dvd_file;
 }
 
-dvd_file_t *DVDOpenFile( dvd_reader_t *dvd, int titlenum,
+dvd_file_t *DVDOpenFile( dvd_reader_t *ctx, int titlenum,
                          dvd_read_domain_t domain )
 {
+  dvd_reader_device_t *dvd = ctx->rd;
   char filename[ MAX_UDF_FILE_NAME_LEN ];
   int do_cache = 0;
 
@@ -928,35 +964,36 @@ dvd_file_t *DVDOpenFile( dvd_reader_t *dvd, int titlenum,
     break;
   case DVD_READ_MENU_VOBS:
     if( dvd->isImageFile ) {
-      return DVDOpenVOBUDF( dvd, titlenum, 1 );
+      return DVDOpenVOBUDF( ctx, titlenum, 1 );
     } else {
-      return DVDOpenVOBPath( dvd, titlenum, 1 );
+      return DVDOpenVOBPath( ctx, titlenum, 1 );
     }
     break;
   case DVD_READ_TITLE_VOBS:
     if( titlenum == 0 ) return NULL;
     if( dvd->isImageFile ) {
-      return DVDOpenVOBUDF( dvd, titlenum, 0 );
+      return DVDOpenVOBUDF( ctx, titlenum, 0 );
     } else {
-      return DVDOpenVOBPath( dvd, titlenum, 0 );
+      return DVDOpenVOBPath( ctx, titlenum, 0 );
     }
     break;
   default:
-    fprintf( stderr, "libdvdread: Invalid domain for file open.\n" );
+    Log1(ctx, "Invalid domain for file open." );
     return NULL;
   }
 
   if( dvd->isImageFile ) {
-    return DVDOpenFileUDF( dvd, filename, do_cache );
+    return DVDOpenFileUDF( ctx, filename, do_cache );
   } else {
-    return DVDOpenFilePath( dvd, filename );
+    return DVDOpenFilePath( ctx, filename );
   }
 }
 
 void DVDCloseFile( dvd_file_t *dvd_file )
 {
-  if( dvd_file && dvd_file->dvd ) {
-    if( !dvd_file->dvd->isImageFile ) {
+  dvd_reader_device_t *dvd = dvd_file->ctx->rd;
+  if( dvd_file && dvd ) {
+    if( !dvd->isImageFile ) {
       int i;
 
       for( i = 0; i < TITLES_MAX; ++i ) {
@@ -1037,7 +1074,7 @@ static int DVDFileStatVOBPath( dvd_reader_t *dvd, int title,
     return -1;
 
   if( stat( full_path, &fileinfo ) < 0 ) {
-    fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
+    Log1(dvd, "Can't stat() %s.", filename );
     return -1;
   }
 
@@ -1053,7 +1090,7 @@ static int DVDFileStatVOBPath( dvd_reader_t *dvd, int title,
         break;
 
       if( stat( full_path, &fileinfo ) < 0 ) {
-        fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
+        Log1(dvd, "Can't stat() %s.", filename );
         break;
       }
 
@@ -1072,9 +1109,10 @@ static int DVDFileStatVOBPath( dvd_reader_t *dvd, int title,
 }
 
 
-int DVDFileStat( dvd_reader_t *dvd, int titlenum,
+int DVDFileStat( dvd_reader_t *reader, int titlenum,
                  dvd_read_domain_t domain, dvd_stat_t *statbuf )
 {
+  dvd_reader_device_t *dvd = reader->rd;
   char filename[ MAX_UDF_FILE_NAME_LEN ];
   struct stat fileinfo;
   uint32_t size;
@@ -1102,9 +1140,9 @@ int DVDFileStat( dvd_reader_t *dvd, int titlenum,
     break;
   case DVD_READ_MENU_VOBS:
     if( dvd->isImageFile )
-      return DVDFileStatVOBUDF( dvd, titlenum, 1, statbuf );
+      return DVDFileStatVOBUDF( reader, titlenum, 1, statbuf );
     else
-      return DVDFileStatVOBPath( dvd, titlenum, 1, statbuf );
+      return DVDFileStatVOBPath( reader, titlenum, 1, statbuf );
 
     break;
   case DVD_READ_TITLE_VOBS:
@@ -1112,19 +1150,19 @@ int DVDFileStat( dvd_reader_t *dvd, int titlenum,
       return -1;
 
     if( dvd->isImageFile )
-      return DVDFileStatVOBUDF( dvd, titlenum, 0, statbuf );
+      return DVDFileStatVOBUDF( reader, titlenum, 0, statbuf );
     else
-      return DVDFileStatVOBPath( dvd, titlenum, 0, statbuf );
+      return DVDFileStatVOBPath( reader, titlenum, 0, statbuf );
 
     break;
   default:
-    fprintf( stderr, "libdvdread: Invalid domain for file stat.\n" );
+    Log1(reader, "Invalid domain for file stat." );
     errno = EINVAL;
     return -1;
   }
 
   if( dvd->isImageFile ) {
-    if( UDFFindFile( dvd, filename, &size ) ) {
+    if( UDFFindFile( reader, filename, &size ) ) {
       statbuf->size = size;
       statbuf->nr_parts = 1;
       statbuf->parts_size[ 0 ] = size;
@@ -1133,9 +1171,9 @@ int DVDFileStat( dvd_reader_t *dvd, int titlenum,
   } else {
     char full_path[ PATH_MAX + 1 ];
 
-    if( findDVDFile( dvd, filename, full_path ) ) {
+    if( findDVDFile( reader, filename, full_path ) ) {
       if( stat( full_path, &fileinfo ) < 0 )
-        fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
+        Log1(reader, "Can't stat() %s.", filename );
       else {
         statbuf->size = fileinfo.st_size;
         statbuf->nr_parts = 1;
@@ -1148,24 +1186,24 @@ int DVDFileStat( dvd_reader_t *dvd, int titlenum,
 }
 
 /* Internal, but used from dvd_udf.c */
-int InternalUDFReadBlocksRaw( const dvd_reader_t *device, uint32_t lb_number,
+int InternalUDFReadBlocksRaw( const dvd_reader_t *ctx, uint32_t lb_number,
                       size_t block_count, unsigned char *data,
                       int encrypted )
 {
   int ret;
 
-  if( !device->dev ) {
-    fprintf( stderr, "libdvdread: Fatal error in block read.\n" );
+  if( !ctx->rd->dev ) {
+    Log0(ctx, "Fatal error in block read." );
     return -1;
   }
 
-  ret = dvdinput_seek( device->dev, (int) lb_number );
+  ret = dvdinput_seek( ctx->rd->dev, (int) lb_number );
   if( ret != (int) lb_number ) {
-    fprintf( stderr, "libdvdread: Can't seek to block %u\n", lb_number );
+    Log1(ctx, "Can't seek to block %u", lb_number );
     return ret;
   }
 
-  ret = dvdinput_read( device->dev, (char *) data,
+  ret = dvdinput_read( ctx->rd->dev, (char *) data,
                        (int) block_count, encrypted );
   return ret;
 }
@@ -1196,7 +1234,7 @@ static int DVDReadBlocksUDF( const dvd_file_t *dvd_file, uint32_t offset,
     return block_count;
   } else {
     /* use dvdinput access */
-    return InternalUDFReadBlocksRaw( dvd_file->dvd, dvd_file->lb_start + offset,
+    return InternalUDFReadBlocksRaw( dvd_file->ctx, dvd_file->lb_start + offset,
                              block_count, data, encrypted );
   }
 }
@@ -1211,6 +1249,7 @@ static int DVDReadBlocksPath( const dvd_file_t *dvd_file, unsigned int offset,
                               size_t block_count, unsigned char *data,
                               int encrypted )
 {
+  const dvd_reader_t *ctx = dvd_file->ctx;
   int i;
   int ret, ret2, off;
 
@@ -1223,8 +1262,7 @@ static int DVDReadBlocksPath( const dvd_file_t *dvd_file, unsigned int offset,
       if( ( offset + block_count ) <= dvd_file->title_sizes[ i ] ) {
         off = dvdinput_seek( dvd_file->title_devs[ i ], (int)offset );
         if( off < 0 || off != (int)offset ) {
-          fprintf( stderr, "libdvdread: Can't seek to block %u\n",
-                   offset );
+          Log1(ctx, "Can't seek to block %u", offset );
           return off < 0 ? off : 0;
         }
         ret = dvdinput_read( dvd_file->title_devs[ i ], data,
@@ -1238,8 +1276,7 @@ static int DVDReadBlocksPath( const dvd_file_t *dvd_file, unsigned int offset,
         /* Read part 1 */
         off = dvdinput_seek( dvd_file->title_devs[ i ], (int)offset );
         if( off < 0 || off != (int)offset ) {
-          fprintf( stderr, "libdvdread: Can't seek to block %u\n",
-                   offset );
+          Log1(ctx, "Can't seek to block %u", offset );
           return off < 0 ? off : 0;
         }
         ret = dvdinput_read( dvd_file->title_devs[ i ], data,
@@ -1255,8 +1292,7 @@ static int DVDReadBlocksPath( const dvd_file_t *dvd_file, unsigned int offset,
         /* Read part 2 */
         off = dvdinput_seek( dvd_file->title_devs[ i + 1 ], 0 );
         if( off < 0 || off != 0 ) {
-          fprintf( stderr, "libdvdread: Can't seek to block %d\n",
-                   0 );
+          Log1(ctx, "Can't seek to block %d", 0 );
           return off < 0 ? off : 0;
         }
         ret2 = dvdinput_read( dvd_file->title_devs[ i + 1 ],
@@ -1279,6 +1315,8 @@ static int DVDReadBlocksPath( const dvd_file_t *dvd_file, unsigned int offset,
 ssize_t DVDReadBlocks( dvd_file_t *dvd_file, int offset,
                        size_t block_count, unsigned char *data )
 {
+  dvd_reader_t *ctx = dvd_file->ctx;
+  dvd_reader_device_t *dvd = ctx->rd;
   int ret;
 
   /* Check arguments. */
@@ -1286,10 +1324,10 @@ ssize_t DVDReadBlocks( dvd_file_t *dvd_file, int offset,
     return -1;
 
   /* Hack, and it will still fail for multiple opens in a threaded app ! */
-  if( dvd_file->dvd->css_title != dvd_file->css_title ) {
-    dvd_file->dvd->css_title = dvd_file->css_title;
-    if( dvd_file->dvd->isImageFile ) {
-      dvdinput_title( dvd_file->dvd->dev, (int)dvd_file->lb_start );
+  if( dvd->css_title != dvd_file->css_title ) {
+      dvd->css_title = dvd_file->css_title;
+    if( dvd->isImageFile ) {
+      dvdinput_title( dvd->dev, (int)dvd_file->lb_start );
     }
     /* Here each vobu has it's own dvdcss handle, so no need to update
     else {
@@ -1297,7 +1335,7 @@ ssize_t DVDReadBlocks( dvd_file_t *dvd_file, int offset,
     }*/
   }
 
-  if( dvd_file->dvd->isImageFile ) {
+  if( dvd->isImageFile ) {
     ret = DVDReadBlocksUDF( dvd_file, (uint32_t)offset,
                             block_count, data, DVDINPUT_READ_DECRYPT );
   } else {
@@ -1323,18 +1361,20 @@ int32_t DVDFileSeek( dvd_file_t *dvd_file, int32_t offset )
 
 int DVDFileSeekForce(dvd_file_t *dvd_file, int offset, int force_size)
 {
+  dvd_reader_t *ctx = dvd_file->ctx;
+  dvd_reader_device_t *dvd = ctx->rd;
   /* Check arguments. */
   if( dvd_file == NULL || offset <= 0 )
       return -1;
 
-  if( dvd_file->dvd->isImageFile ) {
+  if( dvd->isImageFile ) {
     if( force_size < 0 )
       force_size = (offset - 1) / DVD_VIDEO_LB_LEN + 1;
     if( dvd_file->filesize < force_size ) {
       dvd_file->filesize = force_size;
       free(dvd_file->cache);
       dvd_file->cache = NULL;
-      fprintf(stderr, "libdvdread: Ignored size of file indicated in UDF.\n");
+      Log2(ctx, "Ignored size of file indicated in UDF.");
     }
   }
 
@@ -1347,6 +1387,8 @@ int DVDFileSeekForce(dvd_file_t *dvd_file, int offset, int force_size)
 
 ssize_t DVDReadBytes( dvd_file_t *dvd_file, void *data, size_t byte_size )
 {
+  dvd_reader_t *ctx = dvd_file->ctx;
+  dvd_reader_device_t *dvd = ctx->rd;
   unsigned char *secbuf_base, *secbuf;
   unsigned int numsec, seek_sector, seek_byte;
   int ret;
@@ -1363,13 +1405,12 @@ ssize_t DVDReadBytes( dvd_file_t *dvd_file, void *data, size_t byte_size )
 
   secbuf_base = malloc( numsec * DVD_VIDEO_LB_LEN + 2048 );
   if( !secbuf_base ) {
-    fprintf( stderr, "libdvdread: Can't allocate memory "
-             "for file read!\n" );
+    Log0(ctx, "Can't allocate memory for file read" );
     return 0;
   }
   secbuf = (unsigned char *)(((uintptr_t)secbuf_base & ~((uintptr_t)2047)) + 2048);
 
-  if( dvd_file->dvd->isImageFile ) {
+  if( dvd->isImageFile ) {
     ret = DVDReadBlocksUDF( dvd_file, (uint32_t) seek_sector,
                             (size_t) numsec, secbuf, DVDINPUT_NOFLAGS );
   } else {
@@ -1412,8 +1453,7 @@ int DVDDiscID( dvd_reader_t *dvd, unsigned char *discid )
 
   vmg_ifo = ifoOpen( dvd, 0 );
   if( !vmg_ifo ) {
-    fprintf( stderr, "libdvdread: DVDDiscId, failed to "
-      "open VMG IFO!\n" );
+    Log0(dvd, "DVDDiscId, failed to open VMG IFO" );
     return -1;
   }
 
@@ -1435,8 +1475,7 @@ int DVDDiscID( dvd_reader_t *dvd, unsigned char *discid )
 
       if( buffer_base == NULL ) {
           DVDCloseFile( dvd_file );
-          fprintf( stderr, "libdvdread: DVDDiscId, failed to "
-                   "allocate memory for file read!\n" );
+          Log0(dvd, "DVDDiscId, failed to allocate memory for file read" );
           return -1;
       }
 
@@ -1444,8 +1483,8 @@ int DVDDiscID( dvd_reader_t *dvd, unsigned char *discid )
 
       bytes_read = DVDReadBytes( dvd_file, buffer, file_size );
       if( bytes_read != file_size ) {
-          fprintf( stderr, "libdvdread: DVDDiscId read returned %zd bytes"
-                   ", wanted %zd\n", bytes_read, file_size );
+          Log1(dvd, "DVDDiscId read returned %zd bytes"
+                   ", wanted %zd", bytes_read, file_size );
           DVDCloseFile( dvd_file );
           free( buffer_base );
           return -1;
@@ -1467,10 +1506,11 @@ int DVDDiscID( dvd_reader_t *dvd, unsigned char *discid )
 }
 
 
-int DVDISOVolumeInfo( dvd_reader_t *dvd,
+int DVDISOVolumeInfo( dvd_reader_t *ctx,
                       char *volid, unsigned int volid_size,
                       unsigned char *volsetid, unsigned int volsetid_size )
 {
+  dvd_reader_device_t *dvd = ctx->rd;
   unsigned char *buffer, *buffer_base;
   int ret;
 
@@ -1486,17 +1526,17 @@ int DVDISOVolumeInfo( dvd_reader_t *dvd,
   buffer_base = malloc( DVD_VIDEO_LB_LEN + 2048 );
 
   if( buffer_base == NULL ) {
-    fprintf( stderr, "libdvdread: DVDISOVolumeInfo, failed to "
-             "allocate memory for file read!\n" );
+    Log0(ctx, "DVDISOVolumeInfo, failed to "
+             "allocate memory for file read" );
     return -1;
   }
 
   buffer = (unsigned char *)(((uintptr_t)buffer_base & ~((uintptr_t)2047)) + 2048);
 
-  ret = InternalUDFReadBlocksRaw( dvd, 16, 1, buffer, 0 );
+  ret = InternalUDFReadBlocksRaw( ctx, 16, 1, buffer, 0 );
   if( ret != 1 ) {
-    fprintf( stderr, "libdvdread: DVDISOVolumeInfo, failed to "
-             "read ISO9660 Primary Volume Descriptor!\n" );
+    Log0(ctx, "DVDISOVolumeInfo, failed to "
+             "read ISO9660 Primary Volume Descriptor" );
     free( buffer_base );
     return -1;
   }
@@ -1528,28 +1568,28 @@ int DVDISOVolumeInfo( dvd_reader_t *dvd,
 }
 
 
-int DVDUDFVolumeInfo( dvd_reader_t *dvd,
+int DVDUDFVolumeInfo( dvd_reader_t *ctx,
                       char *volid, unsigned int volid_size,
                       unsigned char *volsetid, unsigned int volsetid_size )
 {
   int ret;
   /* Check arguments. */
-  if( dvd == NULL )
+  if( ctx == NULL || ctx->rd == NULL )
     return -1;
 
-  if( dvd->dev == NULL ) {
+  if( ctx->rd->dev == NULL ) {
     /* No block access, so no UDF VolumeSet Identifier */
     return -1;
   }
 
   if( (volid != NULL) && (volid_size > 0) ) {
-    ret = UDFGetVolumeIdentifier(dvd, volid, volid_size);
+    ret = UDFGetVolumeIdentifier(ctx, volid, volid_size);
     if(!ret) {
       return -1;
     }
   }
   if( (volsetid != NULL) && (volsetid_size > 0) ) {
-    ret =  UDFGetVolumeSetIdentifier(dvd, volsetid, volsetid_size);
+    ret =  UDFGetVolumeSetIdentifier(ctx, volsetid, volsetid_size);
     if(!ret) {
       return -1;
     }
